#include "tcp.h"
#include "../htonl.h"
#include "../uthash.h"
#include "../dynarr.h"

#include <stdlib.h>

typedef struct {
    rpc_tcp_conn_new_cb new_conn;
    rpc_tcp_server_t *server;
} rpc_tcp_state_t;

static void
init_tcp_server(rpc_tcp_server_t *s, uv_tcp_t *sock)
{
    s->_conns = NULL;
    s->_sock = sock;
}

static void
init_tcp_state(rpc_tcp_state_t *s, rpc_tcp_conn_new_cb cb,
	       rpc_tcp_server_t *server)
{
    s->new_conn = cb;
    s->server = server;
    return;
}

static void
init_tcp_conn(rpc_tcp_conn_t *c, rpc_tcp_server_t *_server,
	      UT_hash_handle *hh, uv_tcp_t *_client, dynarr_t *buf)
{
    c->_client = _client;
    c->_server = _server;
    c->_hh = hh;
    c->_buf = buf;
    c->_streams = NULL;
    c->stream_id = 0;
}

static void
on_new_connection(uv_stream_t *stream, int status)
{
    rpc_tcp_conn_new_cb cb = ((rpc_tcp_state_t *)stream->data)->new_conn;
    rpc_tcp_server_t *server = ((rpc_tcp_state_t *)stream->data)->server;

    if(status < 0) {
	goto ERR_CB;
    }

    uv_tcp_t *client = (uv_tcp_t*) malloc(sizeof(uv_tcp_t));
    if(client == NULL) {
	status = -1;
	goto ERR_CB;
    }

    uv_tcp_init(stream->loop, client);

    status = uv_accept(stream, (uv_stream_t*) client);
    if(status < 0) {
	goto FREE_CLIENT;
    }

    rpc_tcp_conn_t *conn = malloc(sizeof(rpc_tcp_conn_t));
    if(conn == NULL) goto FREE_CLIENT;

    dynarr_t *buf = malloc(sizeof(dynarr_t));
    if(buf == NULL) { status = -1; goto FREE_CONNECTION; }
    status = dynarr_init(buf, 65536);
    if(status < 0) { goto FREE_BUF; }

    UT_hash_handle *hh = malloc(sizeof(UT_hash_handle));
    if(hh == NULL) { status = -1; goto FREE_BUF; }

    init_tcp_conn(conn, server, hh, client, buf);
    // Keep a map of all connections, indexed by their client socket ptr
    HASH_ADD(_hh, server->_conns, _client, sizeof(void*), conn);

    cb(conn, 0);

    return;
FREE_BUF:
    free(buf);
FREE_CONNECTION:
    free(conn);
FREE_CLIENT:
    free(client);
ERR_CB:
    free(stream->data);
    cb(NULL, status);
}

rpc_tcp_server_t *
rpc_tcp_listen(uv_tcp_t *sock, int backlog, rpc_tcp_conn_new_cb cb)
{
    rpc_tcp_server_t *ret = NULL;

    rpc_tcp_server_t *server = malloc(sizeof(rpc_tcp_server_t));
    if(server == NULL) goto RET; else ret = server;

    init_tcp_server(server, sock);

    rpc_tcp_state_t *state = malloc(sizeof(rpc_tcp_state_t));
    if(state == NULL) { ret = NULL; goto FREE_SERVER; }
    init_tcp_state(state, cb, server);

    sock->data = state;

    int r = uv_listen((uv_stream_t *)sock, backlog, on_new_connection);
    if(r < 0) { ret = NULL; goto FREE_STATE; }

    // `ret' should contain a pointer to the server at this point
    goto RET;
FREE_STATE:
    free(state);
FREE_SERVER:
    free(server);
RET:
    return ret;
}

static void
cleanup_conn(uv_handle_t *h)
{
    rpc_tcp_conn_t *conn = h->data;
    dynarr_release(conn->_buf);
    free(conn->_buf);
    free(conn->_client);
    free(conn->_hh);
    free(conn);
}

int
rpc_tcp_conn_close(rpc_tcp_conn_t *conn, rpc_tcp_conn_close_cb conn_cb,
		   rpc_tcp_stream_close_cb stream_cb)
{
    rpc_tcp_stream_t *stream, *tmp;
    HASH_ITER(_hh, conn->_streams, stream, tmp) {
	HASH_DELETE(_hh, conn->_streams, stream);
	if(stream_cb != NULL) stream_cb(stream, 0);
	free(stream);
	free(stream->_hh);
    }

    HASH_DELETE(_hh, conn->_server->_conns, conn);
    if(conn_cb != NULL) conn_cb(conn, 0);
    conn->_client->data = conn;
    uv_close((uv_handle_t*)conn->_client, cleanup_conn);
    return 0;
}

static void
cleanup_server(uv_handle_t *h)
{
    // All connections, streams, and hashtable should have been closed & freed
    // at this point. This only leaves the struct itself.
    rpc_tcp_server_t *s = h->data;
    free(s);
}

int
rpc_tcp_close(rpc_tcp_server_t *server, rpc_tcp_conn_close_cb conn_cb,
	      rpc_tcp_stream_close_cb stream_cb)
{
    rpc_tcp_conn_t *conn, *tmp;
    HASH_ITER(_hh, server->_conns, conn, tmp) {
	HASH_DELETE(_hh, server->_conns, conn);

	// Copy-paste of rpc_tcp_conn_close(), sans removing itself from parent
	// server
	{
	    rpc_tcp_stream_t *stream, *tmp;
	    HASH_ITER(_hh, conn->_streams, stream, tmp) {
		HASH_DELETE(_hh, conn->_streams, stream);
		if(stream_cb != NULL) stream_cb(stream, 0);
		free(stream);
		free(stream->_hh);
	    }

	    if(conn_cb != NULL) conn_cb(conn, 0);
	    conn->_client->data = conn;
	    uv_close((uv_handle_t*)conn->_client, cleanup_conn);
	}
    }

    // Allocated in `rpc_tcp_listen()' to pass user-provided callbacks, etc.
    free(server->_sock->data);

    server->_sock->data = server;
    uv_close((uv_handle_t*)server->_sock, cleanup_server);
    return 0;
}

typedef struct {
    rpc_tcp_stream_new_cb cb;
    rpc_tcp_stream_t *s;
    int r;
} rpc_new_stream_state_t;

static void
init_tcp_stream(rpc_tcp_stream_t *s,
		uv_stream_t *_stream,
		rpc_tcp_conn_t *_owner,
		UT_hash_handle *_hh,
		uint16_t id)
{
    s->_stream = _stream;
    s->owner = _owner;
    s->id = id;

}

static void
defer_new_stream_cb(uv_timer_t *t)
{
    rpc_new_stream_state_t *state = t->data;
    rpc_tcp_stream_new_cb cb = state->cb;
    rpc_tcp_stream_t *s = state->s;
    int r = state->r;

    uv_close((uv_handle_t *)t, (uv_close_cb)free);
    free(state);

    cb(s, r);
}

int
rpc_tcp_stream_open(rpc_tcp_conn_t *conn, rpc_tcp_stream_new_cb cb)
{
    int status = 0;

    rpc_tcp_stream_t *stream = malloc(sizeof(rpc_tcp_stream_t));
    if(stream == NULL) {
	status = -1;
	goto RET;
    }

    // TODO: keep track of and reuse freed stream ids

    UT_hash_handle *hh = malloc(sizeof(UT_hash_handle));
    if(hh == NULL) { status = -1; goto FREE_STREAM; }

    init_tcp_stream(stream, (uv_stream_t *)conn->_client,
		    conn, hh, conn->stream_id++);
    // Keep track of all streams by their id in a hashmap
    HASH_ADD(_hh, conn->_streams, id, 2, stream);


    // TODO: notify peer of new stream instead of just tagging frames

    uv_timer_t *t = malloc(sizeof(uv_timer_t));
    if(t == NULL) {
	status = -1;
	goto FREE_HH;
    }

    // Yield execution with a next-iteration timer
    uv_timer_init(conn->_server->_sock->loop, t);
    uv_timer_start(t, defer_new_stream_cb, 0, 0);

    goto RET;
FREE_HH:
    free(hh);
FREE_STREAM:
    free(stream);
RET:
    return status;
}

int
rpc_tcp_stream_close(rpc_tcp_stream_t *stream, rpc_tcp_stream_close_cb cb)
{
    // Delete the stream from the owning connection's hashmap
    HASH_DELETE(_hh, stream->owner->_streams, stream);
    if(cb != NULL) cb(stream, 0);
    free(stream->_hh);
    free(stream);
    return 0;
}


typedef struct {
    rpc_tcp_stream_msg_cb cb;
    rpc_tcp_stream_t *stream;
    uv_buf_t *bufs;
} stream_send_state_t;

static void
on_write_complete(uv_write_t *handle, int status)
{
    stream_send_state_t *data = handle->data;
    rpc_tcp_stream_msg_cb cb = data->cb;
    rpc_tcp_stream_t *stream = data->stream;
    uv_buf_t buf = data->bufs[1];

    free(data->bufs[0].base);
    free(data->bufs);
    free(data);
    uv_close((uv_handle_t*)handle, (uv_close_cb)free);

    cb(stream, status, buf.base, buf.len);
}

int
rpc_tcp_stream_send(rpc_tcp_stream_t *stream, void *buf, uint32_t n,
		    rpc_tcp_stream_msg_cb cb)
{
    int status = 0;

    uv_write_t *w = malloc(sizeof(uv_write_t));
    if(w == NULL) {
	status = -1;
	goto RET;
    }

    uv_buf_t *bufs = malloc(sizeof(uv_buf_t) * 2);
    if(bufs == NULL) {
	status = -1;
	goto FREE_REQ;
    }

    rpc_tcp_msg_header *header = malloc(sizeof(rpc_tcp_msg_header));
    if(header == NULL) {
	status = -1;
	goto FREE_BUFS;
    }

    stream_send_state_t *data = malloc(sizeof(stream_send_state_t));
    if(data == NULL) {
	status = -1;
	goto FREE_HEADER;
    }

    data->cb = cb;
    data->stream = stream;
    data->bufs = bufs;

    header->sz = htonl(n);
    header->id = htons(stream->id);

    bufs[0].base = (char *)header;
    bufs[0].len = sizeof(rpc_tcp_msg_header);

    bufs[1].base = buf;
    bufs[1].len = n;

    uv_write(w, stream->_stream, bufs, 2, on_write_complete);

    return 0;
FREE_HEADER:
    free(header);
FREE_BUFS:
    free(bufs);
FREE_REQ:
    free(w);
RET:
    return status;
}

typedef struct {
    rpc_tcp_conn_t *owner;
    size_t total_sz, left_sz;
    uint16_t stream_id;
    unsigned char is_reading_msg;
} read_state;

static void
read_alloc(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
{
    buf->base = malloc(suggested_size);
    buf->len = suggested_size;
}

static void
read_init(read_state *s, void *buf, ssize_t nread);

static void
read_body(read_state *s, void *buf, ssize_t nread)
{
    rpc_tcp_conn_t *c = s->owner;
    ssize_t leftover = s->left_sz - nread;

#define HAND_OFF_MSG							\
    void *message = malloc(s->total_sz);				\
    memcpy(message, c->_buf->data, c->_buf->sz);			\
    dynarr_clear(c->_buf, 0);						\
    if (s->stream_id == 0) {						\
	if(c->_recv != NULL)						\
	    c->_recv(NULL, 0, message, s->total_sz);			\
	else free(message);						\
    }									\
    else if(s->stream_id != 0) {					\
	rpc_tcp_stream_t *stream = NULL;				\
	HASH_FIND(_hh, c->_streams, &s->stream_id, sizeof(s->stream_id), \
		  stream);						\
	if(stream != NULL) {						\
	    if(stream->_recv != NULL)					\
		c->_recv(stream, 0, message, s->total_sz);		\
	    else free(message);						\
	} else free(message);						\
    }

    // We've read the entire message, exactly
    if (leftover == 0) {
	dynarr_append(c->_buf, buf, s->left_sz);
	HAND_OFF_MSG;
    }

    // We've read a chunk of the message
    else if(leftover > 0) {
	dynarr_append(c->_buf, buf, s->left_sz);
    }

    // We've read more than just the current message
    else if (leftover < 0) {
	dynarr_append(c->_buf, buf, s->left_sz);

	HAND_OFF_MSG;

	s->is_reading_msg = 0;
	read_init(s, buf + (nread + leftover), -leftover);
    }
#undef HAND_OFF_MSG
}

static void
read_init(read_state *s, void *buf, ssize_t nread)
{
    if(nread < sizeof(rpc_tcp_msg_header)) {
	// TODO: handle this case
	fprintf(stderr, "read_cb(stub): initial message packet smaller "
		"than header");
	exit(2);
    }

    s->is_reading_msg = 1;
    s->total_sz  = ntohl(*(uint32_t *)(buf + offsetof(rpc_tcp_msg_header, sz)));
    s->stream_id = ntohs(*(uint16_t *)(buf + offsetof(rpc_tcp_msg_header, id)));
    s->left_sz = s->total_sz;

    size_t leftover = nread - sizeof(rpc_tcp_msg_header);
    if(leftover > 0) {
	read_body(s, buf + sizeof(rpc_tcp_msg_header), leftover);
    }
}

static void
read_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
{
    if(nread == 0) goto FREE_BUF;
    if(nread < 0) {
	read_state *s = stream->data;
	s->is_reading_msg = 0;
	dynarr_clear(s->owner->_buf, 0);

	s->owner->_recv((rpc_tcp_stream_t*)s->owner, nread, NULL, -1);

	// Pipe broken: prepare to destroy the connection
	if(nread == UV_EOF) {
	    free(s);
	}

	goto FREE_BUF;
    }

    read_state *s = (read_state *)stream->data;
    rpc_tcp_conn_t *c = s->owner;

    if(s->is_reading_msg) {
	// Already in the middle of reading a message
	read_body(s, buf->base, nread);
    } else {
	// Read the message starting from the header
	read_init(s, buf->base, nread);
    }

FREE_BUF:
    if(buf != NULL && buf->base != NULL)
	free(buf->base);
}

int
rpc_tcp_conn_read_start(rpc_tcp_conn_t *conn, rpc_tcp_stream_msg_cb cb)
{
    read_state *state = malloc(sizeof(read_state));
    if(state == NULL) return -1;
    *state = (read_state){
	.owner = conn,
	.total_sz = 0,
	.left_sz  = 0,
	.is_reading_msg  = 0
    };

    conn->_client->data = state;
    conn->_recv = cb;

    uv_read_start((uv_stream_t*)conn->_client, read_alloc, read_cb);
}

int
rpc_tcp_conn_read_stop(rpc_tcp_conn_t *conn)
{
    conn->_recv = NULL;
    rpc_tcp_conn_read_stop_all(conn);
    free(conn->_client->data);
    return 0;
}

int
rpc_tcp_conn_read_stop_all(rpc_tcp_conn_t *conn)
{
    rpc_tcp_stream_t *stream, *tmp;
    HASH_ITER(_hh, conn->_streams, stream, tmp) {
	stream->_recv = NULL;
    }
}

int
rpc_tcp_stream_read_start(rpc_tcp_stream_t *stream, rpc_tcp_stream_msg_cb cb)
{
    stream->_recv = cb;
    return 0;
}

int
rpc_tcp_stream_read_stop(rpc_tcp_stream_t *stream)
{
    stream->_recv = NULL;
    return 0;
}
