#ifndef RPC_TRANSPORT_H
#define RPC_TRANSPORT_H

#include <uv.h>

#pragma pack(1)
typedef struct {
    uint32_t sz;
    uint16_t id;
} rpc_tcp_msg_header;

// Forward decl for opaque hashtable impl
typedef struct UT_hash_handle UT_hash_handle;
typedef struct rpc_tcp_conn_s rpc_tcp_conn_t;
typedef struct rpc_tcp_server_s rpc_tcp_server_t;
typedef struct rpc_tcp_stream_s rpc_tcp_stream_t;

typedef void(*rpc_tcp_stream_msg_cb)(rpc_tcp_stream_t*, int status,
				     void *buf, ssize_t n);
typedef rpc_tcp_stream_msg_cb rpc_tcp_msg_cb;

typedef struct rpc_tcp_stream_s {
    void *data;
    uint16_t id;
    rpc_tcp_msg_cb _recv;
    uv_stream_t *_stream;
    UT_hash_handle *_hh;
    rpc_tcp_conn_t *owner;
} rpc_tcp_stream_t;

typedef struct dynarr_s dynarr_t;

typedef struct rpc_tcp_conn_s {
    void *data;
    uint16_t stream_id;
    dynarr_t *_buf;
    rpc_tcp_msg_cb _recv;
    UT_hash_handle *_hh;
    rpc_tcp_server_t *_server;
    uv_tcp_t *_client;
    rpc_tcp_stream_t *_streams;
} rpc_tcp_conn_t;

typedef struct rpc_tcp_server_s {
    rpc_tcp_conn_t *_conns;
    uv_tcp_t *_sock;
} rpc_tcp_server_t;

typedef void(*rpc_tcp_server_close_cb)(rpc_tcp_server_t*, int status);

typedef void(*rpc_tcp_conn_new_cb)(rpc_tcp_conn_t*, int status);
typedef void(*rpc_tcp_conn_close_cb)(rpc_tcp_conn_t*, int status);

typedef void(*rpc_tcp_stream_new_cb)(rpc_tcp_stream_t*, int status);
typedef void(*rpc_tcp_stream_close_cb)(rpc_tcp_stream_t*, int status);

/**
 * Upgrade a bound uv_tcp_t socket #sock into an #rpc_tcp_server_t.
 * Callback #cb will be called on each new connection.
 */
rpc_tcp_server_t *
rpc_tcp_listen(uv_tcp_t *sock, int backlog, rpc_tcp_conn_new_cb cb);

/**
 * Closes a TCP server #server.
 *
 * #conn_cb (#stream_cb) will be called when closing each connection (stream),
 * to give you a chance to release any user-allocated data. You should free any
 * user-allocated resources on the server before calling this function.
 *
 * All pointers to the server, and any connection or stream belonging to the
 * server are invalidated after a call to this function.
 */
int
rpc_tcp_close(rpc_tcp_server_t *server, rpc_tcp_conn_close_cb conn_cb,
	      rpc_tcp_stream_close_cb stream_cb);

/**
 * Closes the connection #conn and all its streams.
 *
 * #stream_cb will be called when closing each stream to give you a chance to
 * release any user-allocated data. #conn_cb will be called when closing the
 * connection to give you a chance to release any user-allocated data.
 *
 * The connection will be removed from its server after this call. All pointers
 * to the connection or its streams are invalidated after a call to this
 * function.
 *
 */
int
rpc_tcp_conn_close(rpc_tcp_conn_t *conn, rpc_tcp_conn_close_cb conn_cb,
		   rpc_tcp_stream_close_cb stream_cb);

/**
 * Starts reading for messages on the connection's default stream.
 *
 * Callback stream pointer argument will be NULL. Ownership of the returned
 * buffer (if `status == 0') is transferred to the callback.
 */
int
rpc_tcp_conn_read_start(rpc_tcp_conn_t *conn, rpc_tcp_stream_msg_cb cb);

/**
 * Stops reading for messages on the connection altogether.
 */
int
rpc_tcp_conn_read_stop(rpc_tcp_conn_t *conn);

/**
 * Stops reading for messages on all of the connection's streams (but not the
 * connection itself).
 */
int
rpc_tcp_conn_read_stop_all(rpc_tcp_conn_t *conn);

/**
 * Opens a stream on this connection. #cb will be called with the opened stream
 * if successful, or `status < 0' otherwise.
 */
int
rpc_tcp_stream_open(rpc_tcp_conn_t *conn, rpc_tcp_stream_new_cb cb);

/**
 * Closes the stream #stream. #cb will called when closing the stream to give
 * you a change to release any user-allocated data.
 *
 * The stream will be removed from its connection after this call. All pointers
 * to the stream are invalidated after a call to this function.
 */
int
rpc_tcp_stream_close(rpc_tcp_stream_t *stream, rpc_tcp_stream_close_cb cb);

/**
 * Send message #buf of length #n over stream #stream. Calls #cb after
 * finishing.
 *
 * The provided buffer and length will be returned untouched to the callback,
 * with `status == 0' if successful.
 */
int
rpc_tcp_stream_send(rpc_tcp_stream_t *stream, void *buf, uint32_t n,
		    rpc_tcp_stream_msg_cb cb);

/**
 * Start listening to messages on #stream with #cb.
 */
int
rpc_tcp_stream_read_start(rpc_tcp_stream_t *stream, rpc_tcp_stream_msg_cb cb);

/**
 * Stop listening to messages on #stream.
 */
int
rpc_tcp_stream_read_stop(rpc_tcp_stream_t *stream);


#endif //RPC_TRANSPORT_H
